"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("reflect-metadata");
let __optional = Symbol('optional');
let __keys = Symbol('keys');
let __mapping = Symbol('mapping');
let __union = Symbol('union');
let __code = Symbol('code');
let __gt = Symbol('gt');
let __lt = Symbol('lt');
let __eq = Symbol('eq');
let __ne = Symbol('ne');
let __gte = Symbol('gte');
let __lte = Symbol('lte');
let __passthrough = Symbol('passthrough');
class JSONObject extends Object {
    set(key, value) {
        let new_value;
        if (typeof value == 'object') {
            new_value = new (Reflect.getMetadata("design:type", this, key))(value);
        }
        else {
            if (Reflect.hasMetadata(__union, this, key)) {
                let values = (Reflect.getMetadata(__union, this, key));
                let is_valid = false;
                for (let i = 0; i < values.length && !is_valid; ++i) {
                    if (value === values[i]) {
                        is_valid = true;
                    }
                }
                if (!is_valid) {
                    throw new TypeError(`${this.constructor.name}.${key} requires one of the following values:[${values}], got '${value}' instead`);
                }
            }
            if (Reflect.hasMetadata(__passthrough, this, key) && Reflect.getMetadata(__passthrough, this, key) == true) {
                new_value = value;
            }
            else {
                let expected = Reflect.getMetadata("design:type", this, key)(value);
                let expected_type = typeof expected;
                let value_type = typeof value;
                if (expected_type == value_type) {
                    new_value = Reflect.getMetadata("design:type", this, key)(value);
                }
                else {
                    throw new TypeError(`${this.constructor.name}.${key} requires type '${expected_type}', got '${value_type}' instead`);
                }
            }
            let symbols = [__code, __gt, __gte, __eq, __ne, __lt, __lte];
            for (let symbol of symbols) {
                if (Reflect.hasMetadata(symbol, this, key)) {
                    let code = (Reflect.getMetadata(symbol, this, key));
                    code(this, key, new_value);
                }
            }
        }
        this[key] = new_value;
    }
    get(key) {
        return this[key];
    }
    constructor(json) {
        super();
        let json_keys = JSONObject.jsonKeys(this);
        if (!json_keys)
            return; // no properties defined
        for (let key of JSONObject.jsonKeys(this)) {
            let json_value;
            let json_key;
            // Is there a mapping this value?
            if (Reflect.hasMetadata(__mapping, this, key)) {
                json_key = Reflect.getMetadata(__mapping, this, key);
            }
            else {
                json_key = key;
            }
            json_value = json[json_key];
            if (json_value === undefined) {
                if (this.get(key) === undefined && !JSONObject.isOptional(this, key)) {
                    throw new TypeError(`${this.constructor.name}.${key} is required`);
                }
            }
            else {
                this.set(key, json_value);
            }
        }
    }
    static validate(code) {
        return Reflect.metadata(__code, code);
    }
    static union(values) {
        return Reflect.metadata(__union, values);
    }
    static map(newKey) {
        return Reflect.metadata(__mapping, newKey);
    }
    static isOptional(target, key) {
        let ret = Reflect.getMetadata(__optional, target, key);
        return ret == true || ret === undefined;
    }
    static jsonKeys(target) {
        return Reflect.getMetadata(__keys, target);
    }
    static preprocess(target, key) {
        let optional = Reflect.getMetadata(__optional, target, key);
        if (optional !== undefined) {
            throw new SyntaxError(`Cannot mark ${target.constructor.name} as required, already set as ${optional ? 'optional' : 'required'}`);
        }
        let keys;
        if (Reflect.hasMetadata(__keys, target)) {
            keys = Reflect.getMetadata(__keys, target);
        }
        else {
            keys = Array();
            Reflect.defineMetadata(__keys, keys, target);
        }
        keys.push(key);
    }
    validateNumeric(key, value) {
        if (typeof value !== 'number') {
            throw new TypeError(`${this.constructor.name}.${key}: requires numeric type for comparison operators`);
        }
    }
    static gt(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue > capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} > ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__gt, validator);
    }
    static gte(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue >= capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} >= ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__gte, validator);
    }
    static lt(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue < capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} < ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__lt, validator);
    }
    static lte(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue <= capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} <= ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__lte, validator);
    }
    static eq(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            if (!(jsonValue == capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} == ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__eq, validator);
    }
    static ne(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            if (!(jsonValue != capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} != ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__ne, validator);
    }
    static passthrough(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__passthrough, true, target, key);
    }
    static required(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__optional, false, target, key);
    }
    static optional(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__optional, true, target, key);
    }
}
exports.JSONObject = JSONObject;
function validate(code) {
    return JSONObject.validate(code);
}
exports.validate = validate;
function union(values) {
    return JSONObject.union(values);
}
exports.union = union;
function required(target, key) {
    JSONObject.required(target, key);
}
exports.required = required;
function optional(target, key) {
    JSONObject.optional(target, key);
}
exports.optional = optional;
function map(jsonKey) {
    return JSONObject.map(jsonKey);
}
exports.map = map;
function passthrough(target, key) {
    return JSONObject.passthrough(target, key);
}
exports.passthrough = passthrough;
function gt(value) {
    return JSONObject.gt(value);
}
exports.gt = gt;
function gte(value) {
    return JSONObject.gte(value);
}
exports.gte = gte;
function lt(value) {
    return JSONObject.lt(value);
}
exports.lt = lt;
function lte(value) {
    return JSONObject.lte(value);
}
exports.lte = lte;
function eq(value) {
    return JSONObject.eq(value);
}
exports.eq = eq;
function ne(value) {
    return JSONObject.ne(value);
}
exports.ne = ne;
//# sourceMappingURL=json_object.js.map